function [E,M,O,D,NMS]=canny(I,sigma,p_low,p_high)

%% Set default values of input arguments
if nargin<2
    sigma=.5;
end
if nargin<3
    p_low=30;
end
if nargin<4
    p_high=60;
end

%% Verify that input arguments are valid
 % minimum allowed sigma is 0.5 since it leads to smallest possible Gussian 
 % mask of size 1x3
if sigma<0.5
    sigma=0.5;
end
%percentiles thresholds should be percentages between 0 and 100
if (p_low<0 || p_low>100)
    p_low=30;
    disp('3rd input argument must be a percentage value between 0 and 100.');
    disp('Setting low percentile to 30.');
end
if (p_high<0 || p_high>100)
    p_high=60;
    disp('4th input argument must be a percentage value between 0 and 100.');
    disp('Setting high percentile to 60.');
end
%high percentile should be greater than low percentile
if p_high<p_low
    tmp=p_high;
    p_high=p_low;
    p_low=tmp;
    disp('Low percentile must be less than high percentile.');
    disp('Swapping low and high percentiles');
end    

% Convert to grayscale in case of color image
if ndims(I)==3
    I = rgb2gray(I);
end
% Convert to double before performing any operations on the image
I = double(I);
% Normalize between 0 and 1
I = myAffineRescaling(I,1);

%% Set global variables to be used in Hysteresis Thresholding
 % Otherwise these variables will need to be passed into every recursive
 % call which will lead to a high recursion overhead.
global E NMS t_low

%% Kernels for derivative filtering
 % x is the vertically downwards direction
 % y is the horizontal direction from left to right
 % We use central difference to approximate the 1st-order derivative
mask_x=[1;0;-1]/2;

%% Form a Guassian mask for smoothing in 1D
 % Gaussian curve extends from -inf to inf. So we cut it off so as to have
 % a finite portion of the curve covering 99.7% of the area under the curve.
 % 99.7% of the area under Gaussian curve lies within +-3 standard
 % deviations.
if sigma<0.5
    sigma=0.5;  %minimum allowed sigma is 0.5 since it leads to smallest possible mask size 1x3 
end
sHalf = 3*ceil(sigma);
gMask = 1/(sigma*sqrt(2*pi)) * exp(-((-sHalf:sHalf).^2)./(2*sigma^2));  %+-3 standard deviations

%% Smooth the image
 % Since convolution is a separable operation, we can convolve first in
 % y-direction and then convolve the result in x-direction. This can be
 % done using only a 1D mask instead of a 2D mask. Separate 1D convolutions
 % are cheaper than a single 2D convolution.
Ismoothy=myConvolution_ReflectingBoundaries(I,gMask);          %convolve in y-direction
Ismooth=myConvolution_ReflectingBoundaries(Ismoothy,gMask');   %convolve result in x-direction

%% Compute derivatives
Ix=myConvolution_ReflectingBoundaries(Ismooth,mask_x);
Iy=myConvolution_ReflectingBoundaries(Ismooth,mask_x');

%% Compute gradient magnitudes and orientations
 % All angles are computed from the positive x-axis (left image boundary)
 % +ve angle is counter-clockwise from x-axis
 % -ve angle is clockwise from x-axis 
M=sqrt(Ix.^2+Iy.^2);
O=atan2(Iy,Ix);

%% Quantize orientations into 4 directions
 % Direction 1 is vertical |
 % Direction 2 is positive diagonal \
 % Direction 3 is horizontal __
 % Direction 4 is negative diagonal /
D=zeros(size(O));
D( (O>-pi/8 & O<=pi/8) | (O>7*pi/8 | O<=-7*pi/8) ) = 1;     %vertical direction |
D( (O>pi/8 & O<=3*pi/8) | (O>-7*pi/8 & O<=-5*pi/8) ) = 2;   %positive diagonal direction \
D(      ------------ADD_CODE_HERE------------        ) = 3; %horizontal direction __
D( (O>5*pi/8 & O<=7*pi/8) | (O>-3*pi/8 & O<=-pi/8) ) = 4;   %negative diagonal direction /
D(M==0)=0;  %do not assign any direction for flat regions (i.e., zero gradient magnitude)

%% Perform Non-Maxima-Suppresion (NMS) on the gradient magnitudes
 % Use the quantized orientations to determine which neighbours to check 
NMS=M;
for i=2:size(I,1)-1
    for j=2:size(I,2)-1
        switch D(i,j)
            case 1
                if M(i,j)<M(i-1,j) || M(i,j)<M(i+1,j)
                    NMS(i,j)=0;
                end
            case 2
                if ------------ADD_CODE_HERE------------
                    NMS(i,j)=0;
                end
            case 3
                if ------------ADD_CODE_HERE------------
                    NMS(i,j)=0;
                end
            case 4
                if M(i,j)<M(i-1,j+1) || M(i,j)<M(i+1,j-1)
                    NMS(i,j)=0;
                end
        end
    end
end

%% Set thresholds for Hysteresis Thresholding
 % Set adaptive thresholds via percentiles
 % t_low = p_low percentile of gradient magnitudes after NMS (ignore 0 magnitude pixels)
 % t_high = p_high percentile of gradient magnitudes after NMS (ignore 0 magnitude pixels)
thresholds=prctile(NMS(NMS>0),[p_low p_high]);
t_low=thresholds(1);
t_high=thresholds(2);

%% Perform Hysteresis Thresholding
rlimit=get(0,'RecursionLimit');
set(0,'RecursionLimit',5000);
E=zeros(size(NMS));
for i=1:size(NMS,1)
    for j=1:size(NMS,2)
        if (NMS(i,j)>=t_high && E(i,j)==0)  %if unmarked edge pixel
            E(i,j)=1;
            %imagesc(E); drawnow;
            check_neighbours(i,j);
        end
    end
end
set(0,'RecursionLimit',rlimit);

%% Display and save result
result=[[myAffineRescaling(I,1) myAffineRescaling(M,1) myAffineRescaling(O,1)];...
    [myAffineRescaling(D,1) myAffineRescaling(NMS,1) myAffineRescaling(E,1)]];
imagesc(result); colormap gray;
imwrite(result,'canny_result.png');
end

function check_neighbours(i,j)
global E NMS t_low
for k=i-1:i+1
    for l=j-1:j+1
        if k==i && l==j %do not recheck current pixel
            continue
        end
        if (k<1 || k>size(NMS,1) || l<1 || l>size(NMS,2))   %handle image boundaries
            continue
        else
            if ------------ADD_CODE_HERE------------ && E(k,l)==0
                E(k,l)=1;
                imagesc(E); drawnow;
                ------------ADD_CODE_HERE------------
            end
        end
    end
end
end