function [r,c]=detect_corners(I,sigma,rho,p,method)
% 'I' is the input image
% 'sigma' is the standard deviation of the Gaussian mask for smoothing I
% 'rho' is the standard deviation of the Gaussian mask for smoothing
%       gradients for the structure tensor
% 'p' is the percentile threshold for setting adaptive thresholds on the
%       corner strength
% 'method' can be 'tomasi', 'rohr' or 'harris'

%% Set default values of input arguments
if nargin<2
    % For smoothing the input image to reduce noise that affects gradient
    % computation
    sigma=.5;
end
if nargin<3
    % For smoothing the gradients to integrate/gather gradient 
    % information around a pixel.
    % Gradient smoothing should be more than image smoothing.
    rho=1.5*sigma;
end
if nargin<4
    % Percentile threshold for setting adaptive thresholds on corner strength
    p=90;
end
if nargin<5
    method='harris';
end

%% Verify that input arguments are valid
 % minimum allowed sigma is 0.5 since it leads to smallest possible Gussian 
 % mask of size 1x3
if sigma<0.5
    sigma=0.5;
end
if rho<sigma
    % Gradient smoothing should be more than image smoothing.
    disp('Gradient smoothing should be more than image smoothing.');
    disp('Setting rho equal to 1.5 times sigma.');
    rho=1.5*sigma;
end
%percentile threshold p should be a percentage between 0 and 100
if (p<0 || p>100)
    p=90;
    disp('3rd input argument must be a percentage value between 0 and 100.');
    disp('Setting percentile to 90.');
end

%% Convert to grayscale in case of color image
if ndims(I)==3
    I = rgb2gray(I);
end
%% Convert to double before performing any operations on the image
I = double(I);
%% Normalize between 0 and 1
I = myAffineRescaling(I,1);

%% Kernels for derivative filtering
 % x is the vertically downwards direction
 % y is the horizontal direction from left to right
 % We use central difference to approximate the 1st-order derivative
mask_x=[1;0;-1]/2;

%% Form a Guassian mask for smoothing in 1D
 % Gaussian curve extends from -inf to inf. So we cut it off so as to have
 % a finite portion of the curve covering 99.7% of the area under the curve.
 % 99.7% of the area under Gaussian curve lies within +-3 standard
 % deviations.
sHalf = 3*ceil(sigma);
gMask = 1/(sigma*sqrt(2*pi)) * exp(-((-sHalf:sHalf).^2)./(2*sigma^2));  %+-3 standard deviations

%% Smooth the image
 % Since convolution is a separable operation, we can convolve first in
 % y-direction and then convolve the result in x-direction. This can be
 % done using only a 1D mask instead of a 2D mask. Separate 1D convolutions
 % are cheaper than a single 2D convolution.
Ismoothy=myConvolution_ReflectingBoundaries(I,gMask);          %convolve in y-direction
Ismooth=myConvolution_ReflectingBoundaries(Ismoothy,gMask');   %convolve result in x-direction

%% Compute derivatives
Ix=myConvolution_ReflectingBoundaries(Ismooth,mask_x);
Iy=myConvolution_ReflectingBoundaries(Ismooth,mask_x');

%% Pre-compute values for the structure tensor to save computations
IxIx=Ix.^2;
IyIy=Iy.^2;
IxIy=Ix.*Iy;

%% Form a Guassian mask for smoothing the derivatives in 1D
 % Gaussian curve extends from -inf to inf. So we cut it off so as to have
 % a finite portion of the curve covering 99.7% of the area under the curve.
 % 99.7% of the area under Gaussian curve lies within +-3 standard
 % deviations.
sHalf = 3*ceil(rho);
gMask = 1/(rho*sqrt(2*pi)) * exp(-((-sHalf:sHalf).^2)./(2*rho^2));  %+-3 standard deviations

%% Pre-smooth values for the structure tensor to save computations
IxIx_smooth = myConvolution_ReflectingBoundaries(...
    myConvolution_ReflectingBoundaries(IxIx,gMask),gMask');
IyIy_smooth = myConvolution_ReflectingBoundaries(...
    myConvolution_ReflectingBoundaries(IyIy,gMask),gMask');
IxIy_smooth = myConvolution_ReflectingBoundaries(...
    myConvolution_ReflectingBoundaries(IxIy,gMask),gMask');

% imagesc(I); colormap gray
% axis equal; axis tight; axis off;
% hold on;    %draw subsequent image commands on the figure. do not overwrite.

evec_small=[];
evec_large=[];
skp=1;
corner_strength=zeros(size(I));
if strcmp(method,'harris')
    detJ=zeros(size(I));
end
for i=1:skp:size(I,1)
    for j=1:skp:size(I,2)
        %form the 2x2 structure tensor at location i,j
        J = ------------ADD_CODE_HERE------------
        switch method
            case 'tomasi'
                [v,d] = eigs(J);
                %corner strength = smaller eigenvalue of structure tensor
                corner_strength(i,j) = ------------ADD_CODE_HERE------------
                %v = v*d;
                %evec_large = [evec_large;[j,i,v(1,1),v(2,1)]];
                %evec_small = [evec_small;[j,i,v(1,2),v(2,2)]];
            case 'rohr'
                %corner strength = determinant of structure tensor
                corner_strength(i,j) = ------------ADD_CODE_HERE------------
            case 'harris'
                detJ(i,j) = J(1,1)*J(2,2)-J(1,2)^2;
                %corner strength = trace of structure tensor
                corner_strength(i,j) = ------------ADD_CODE_HERE------------
        end
    end
end

if strcmp(method,'tomasi')
    %% draw eigenvectors (large in blue, small in red)
    %quiver(evec_large(:,1),evec_large(:,2),evec_large(:,3),evec_large(:,4),1,'b','LineWidth',2);
    %quiver(evec_small(:,1),evec_small(:,2),evec_small(:,3),evec_small(:,4),1,'r','LineWidth',2);
end

if strcmp(method,'harris')
    local_maxima = find_local_maxima(detJ./corner_strength);
else
    local_maxima = find_local_maxima(corner_strength);
end

%% set adaptive threshold for corner strength via percentile (ignore zero strength corners)
th = prctile(corner_strength(corner_strength>0),p);  %find p_th percentile

%% find corners by thresholding the locally maximum corner strengths
[r,c]=find(local_maxima & corner_strength>=th);